-module(gulid).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([new/0, new_monotonic/1, from_bitarray/1, from_string_function/0, to_parts/1, from_parts/2, from_tuple/1, to_bitarray/1, to_string_function/0, new_as_string/0]).
-export_type([ulid/0, ulid_error/0, array/1]).

-opaque ulid() :: {ulid, bitstring()}.

-type ulid_error() :: {decode_error, binary()} | {invalid_length, binary()}.

-type array(GLH) :: any() | {gleam_phantom, GLH}.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 73).
-spec new() -> ulid().
new() ->
    Time = os:system_time(millisecond),
    Randomness = crypto:strong_rand_bytes(10),
    {ulid, <<Time:48/big, Randomness/bitstring>>}.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 84).
-spec new_monotonic(ulid()) -> ulid().
new_monotonic(Prev_ulid) ->
    Time = os:system_time(millisecond),
    {ulid, <<Prev_time:48/unsigned, Random:80/unsigned>>} = case Prev_ulid of
        {ulid, <<_:48/unsigned, _:80/unsigned>>} -> Prev_ulid;
        _assert_fail ->
            erlang:error(#{gleam_error => let_assert,
                        message => <<"Pattern match failed, no pattern matched the value."/utf8>>,
                        value => _assert_fail,
                        module => <<"gulid"/utf8>>,
                        function => <<"new_monotonic"/utf8>>,
                        line => 86})
    end,
    case gleam@int:compare(Time, Prev_time) of
        eq ->
            {ulid, <<Prev_time:48/big, ((Random + 1)):80/big>>};

        lt ->
            {ulid, <<Prev_time:48/big, ((Random + 1)):80/big>>};

        _ ->
            {ulid, <<Time:48/big, (crypto:strong_rand_bytes(10))/bitstring>>}
    end.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 101).
-spec from_bitarray(bitstring()) -> {ok, ulid()} | {error, ulid_error()}.
from_bitarray(Array) ->
    case Array of
        <<Ulid:128/bitstring>> ->
            {ok, {ulid, Ulid}};

        _ ->
            _pipe = <<"Invalid length of a Ulid bitarray"/utf8>>,
            _pipe@1 = {invalid_length, _pipe},
            {error, _pipe@1}
    end.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 122).
-spec from_string_function() -> fun((binary()) -> {ok, ulid()} |
    {error, ulid_error()}).
from_string_function() ->
    Base32_decode = maps:from_list(
        [{<<"0"/utf8>>, 0},
            {<<"o"/utf8>>, 0},
            {<<"O"/utf8>>, 0},
            {<<"1"/utf8>>, 1},
            {<<"l"/utf8>>, 1},
            {<<"L"/utf8>>, 1},
            {<<"I"/utf8>>, 1},
            {<<"i"/utf8>>, 1},
            {<<"2"/utf8>>, 2},
            {<<"3"/utf8>>, 3},
            {<<"4"/utf8>>, 4},
            {<<"5"/utf8>>, 5},
            {<<"6"/utf8>>, 6},
            {<<"7"/utf8>>, 7},
            {<<"8"/utf8>>, 8},
            {<<"9"/utf8>>, 9},
            {<<"a"/utf8>>, 10},
            {<<"A"/utf8>>, 10},
            {<<"b"/utf8>>, 11},
            {<<"B"/utf8>>, 11},
            {<<"c"/utf8>>, 12},
            {<<"C"/utf8>>, 12},
            {<<"d"/utf8>>, 13},
            {<<"D"/utf8>>, 13},
            {<<"e"/utf8>>, 14},
            {<<"E"/utf8>>, 14},
            {<<"f"/utf8>>, 15},
            {<<"F"/utf8>>, 15},
            {<<"g"/utf8>>, 16},
            {<<"G"/utf8>>, 16},
            {<<"h"/utf8>>, 17},
            {<<"H"/utf8>>, 17},
            {<<"j"/utf8>>, 18},
            {<<"J"/utf8>>, 18},
            {<<"k"/utf8>>, 19},
            {<<"K"/utf8>>, 19},
            {<<"m"/utf8>>, 20},
            {<<"M"/utf8>>, 20},
            {<<"n"/utf8>>, 21},
            {<<"N"/utf8>>, 21},
            {<<"p"/utf8>>, 22},
            {<<"P"/utf8>>, 22},
            {<<"q"/utf8>>, 23},
            {<<"Q"/utf8>>, 23},
            {<<"r"/utf8>>, 24},
            {<<"R"/utf8>>, 24},
            {<<"s"/utf8>>, 25},
            {<<"S"/utf8>>, 25},
            {<<"t"/utf8>>, 26},
            {<<"T"/utf8>>, 26},
            {<<"v"/utf8>>, 27},
            {<<"V"/utf8>>, 27},
            {<<"w"/utf8>>, 28},
            {<<"W"/utf8>>, 28},
            {<<"x"/utf8>>, 29},
            {<<"X"/utf8>>, 29},
            {<<"y"/utf8>>, 30},
            {<<"Y"/utf8>>, 30},
            {<<"z"/utf8>>, 31},
            {<<"Z"/utf8>>, 31}]
    ),
    fun(Str) -> _pipe = Str,
        _pipe@1 = gleam@string:to_graphemes(_pipe),
        _pipe@4 = gleam@list:fold(
            _pipe@1,
            {ok, <<>>},
            fun(Acc, Symb) -> case Acc of
                    {ok, Ulid_bits} ->
                        case gleam@dict:get(Base32_decode, Symb) of
                            {ok, Val} ->
                                {ok,
                                    gleam@bit_array:append(
                                        Ulid_bits,
                                        <<Val:5/big>>
                                    )};

                            {error, _} ->
                                _pipe@2 = (<<<<"Symbol "/utf8, Symb/binary>>/binary,
                                    " is invalid in Base32 encoding"/utf8>>),
                                _pipe@3 = {decode_error, _pipe@2},
                                {error, _pipe@3}
                        end;

                    _ ->
                        Acc
                end end
        ),
        _pipe@5 = gleam@result:map(
            _pipe@4,
            fun(Ulid_bits@1) -> case Ulid_bits@1 of
                    <<_:2, Trim:128/bitstring>> ->
                        Trim;

                    _ ->
                        Ulid_bits@1
                end end
        ),
        gleam@result:'try'(_pipe@5, fun from_bitarray/1) end.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 226).
-spec to_parts(ulid()) -> {integer(), integer()}.
to_parts(Ulid) ->
    {ulid, <<Timestamp:48/unsigned, Random:80/unsigned>>} = case Ulid of
        {ulid, <<_:48/unsigned, _:80/unsigned>>} -> Ulid;
        _assert_fail ->
            erlang:error(#{gleam_error => let_assert,
                        message => <<"Pattern match failed, no pattern matched the value."/utf8>>,
                        value => _assert_fail,
                        module => <<"gulid"/utf8>>,
                        function => <<"to_parts"/utf8>>,
                        line => 227})
    end,
    {Timestamp, Random}.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 233).
-spec from_parts(integer(), integer()) -> ulid().
from_parts(Timestamp, Random) ->
    {ulid, <<Timestamp:48/big, Random:80/big>>}.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 238).
-spec from_tuple({integer(), integer()}) -> ulid().
from_tuple(Parts) ->
    from_parts(erlang:element(1, Parts), erlang:element(2, Parts)).

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 243).
-spec to_bitarray(ulid()) -> bitstring().
to_bitarray(Ulid) ->
    {ulid, Result} = Ulid,
    <<Result/bitstring>>.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 267).
-spec encode_to_string_with_accumulator(bitstring(), binary(), array(binary())) -> {ok,
        binary()} |
    {error, ulid_error()}.
encode_to_string_with_accumulator(Array, Acc, Base32) ->
    case Array of
        <<>> ->
            {ok, Acc};

        <<Symb:5/unsigned, Rest/bitstring>> ->
            _pipe = Acc,
            _pipe@1 = gleam@string:append(_pipe, array:get(Symb, Base32)),
            encode_to_string_with_accumulator(Rest, _pipe@1, Base32);

        _ ->
            _pipe@2 = <<"Given bit array's bit length is not divisible by 5"/utf8>>,
            _pipe@3 = {invalid_length, _pipe@2},
            {error, _pipe@3}
    end.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 286).
-spec to_bitarray_130bits(ulid()) -> bitstring().
to_bitarray_130bits(Ulid) ->
    <<0:2, (to_bitarray(Ulid))/bitstring>>.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 54).
-spec to_string_function() -> fun((ulid()) -> binary()).
to_string_function() ->
    Base32_array = begin
        _pipe = <<"0123456789ABCDEFGHJKMNPQRSTVWXYZ"/utf8>>,
        _pipe@1 = gleam@string:to_graphemes(_pipe),
        array:from_list(_pipe@1)
    end,
    fun(Ulid) ->
        _assert_subject = begin
            _pipe@2 = Ulid,
            _pipe@3 = to_bitarray_130bits(_pipe@2),
            encode_to_string_with_accumulator(
                _pipe@3,
                <<""/utf8>>,
                Base32_array
            )
        end,
        {ok, Ulid_str} = case _assert_subject of
            {ok, _} -> _assert_subject;
            _assert_fail ->
                erlang:error(#{gleam_error => let_assert,
                            message => <<"Pattern match failed, no pattern matched the value."/utf8>>,
                            value => _assert_fail,
                            module => <<"gulid"/utf8>>,
                            function => <<"to_string_function"/utf8>>,
                            line => 63})
        end,
        Ulid_str
    end.

-file("/Users/svava/Projects/Gleam/ulid/src/gulid.gleam", 95).
-spec new_as_string() -> binary().
new_as_string() ->
    _pipe = new(),
    (to_string_function())(_pipe).
